#include <iostream>

#include "payoff.h"
#include "option.h"
#include "correlated_snd.h"
#include "heston_mc.h"

void generate_normal_correlation_paths(DCO_T<double> rho, 
    std::vector<DCO_T<double>>& spot_normals, std::vector<DCO_T<double>>& cor_normals) {
  unsigned vals = spot_normals.size();

  // Create the Standard Normal Distribution and random draw vectors
  StandardNormalDistribution snd;
  std::vector<DCO_T<double>> snd_uniform_draws(vals, 0.0);

  // Simple random number generation method based on RAND
  for (int i=0; i<snd_uniform_draws.size(); i++) {
    snd_uniform_draws[i] = rand() / static_cast<DCO_T<double>>(RAND_MAX);
  }

  // Create standard normal random draws
  snd.random_draws(snd_uniform_draws, spot_normals);

  // Create the correlated standard normal distribution
  CorrelatedSND csnd(rho, &spot_normals);
  std::vector<DCO_T<double>> csnd_uniform_draws(vals, 0.0);

  // Uniform generation for the correlated SND
  for (int i=0; i<csnd_uniform_draws.size(); i++) {
    csnd_uniform_draws[i] = rand() / static_cast<DCO_T<double>>(RAND_MAX);
  }

  // Now create the -correlated- standard normal draw series
  csnd.random_draws(csnd_uniform_draws, cor_normals);
}

int main(int argc, char **argv) {
  // First we create the parameter list
  // Note that you could easily modify this code to input the parameters
  // either from the command line or via a file
  unsigned num_sims = 100000;   // Number of simulated asset paths
  unsigned num_intervals = 100;  // Number of intervals for the asset path to be sampled 

  DCO_T<double> S_0 = 100.0;    // Initial spot price
  DCO_T<double> K = 100.0;      // Strike price
  DCO_T<double> r = 0.0319;     // Risk-free rate
  DCO_T<double> v_0 = 0.010201; // Initial volatility 
  DCO_T<double> T = 1.00;       // One year until expiry

  DCO_T<double> rho = -0.7;     // Correlation of asset and volatility
  DCO_T<double> kappa = 6.21;   // Mean-reversion rate
  DCO_T<double> theta = 0.019;  // Long run average volatility
  DCO_T<double> xi = 0.61;      // "Vol of vol"

  std::vector<DCO_T<double>*> X={&S_0,&K,&r,&v_0,&T,&rho,&kappa,&theta,&xi};
for (const auto& x:X) {
  srand(0); // generate compatible random numbers 
  dco::derivative(*x)=1;

  // Create the PayOff, Option and Heston objects
  PayOff* pPayOffCall = new PayOffCall(K);
  Option* pOption = new Option(K, r, T, pPayOffCall);
  HestonEuler hest_euler(pOption, kappa, theta, xi, rho);

  // Create the spot and vol initial normal and price paths
  std::vector<DCO_T<double>> spot_draws(num_intervals, 0.0);  // Vector of initial spot normal draws
  std::vector<DCO_T<double>> vol_draws(num_intervals, 0.0);   // Vector of initial correlated vol normal draws
  std::vector<DCO_T<double>> spot_prices(num_intervals, S_0);  // Vector of initial spot prices
  std::vector<DCO_T<double>> vol_prices(num_intervals, v_0);   // Vector of initial vol prices

  // Monte Carlo options pricing
  DCO_T<double> payoff_sum = 0.0;
  for (unsigned i=0; i<num_sims; i++) {
    generate_normal_correlation_paths(rho, spot_draws, vol_draws);
    hest_euler.calc_vol_path(vol_draws, vol_prices);
    hest_euler.calc_spot_path(spot_draws, vol_prices, spot_prices);
    payoff_sum += pOption->pay_off->operator()(spot_prices[num_intervals-1]);
  }
  DCO_T<double> option_price = (payoff_sum / static_cast<DCO_T<double>>(num_sims)) * exp(-r*T);
  if (x==&S_0) {
    std::cout << "Option Price: " << option_price << std::endl;
    std::cout << "d Option Price / d X: " << std::endl;
  }
  std::cout << dco::derivative(option_price) << std::endl;
  dco::derivative(*x)=0;

  // Free memory
  delete pOption;
  delete pPayOffCall;
}

  return 0;
}
